open Absyn;;
open Flags;;
open Solve;;
module P = Perm;;
open Printer;;
open Util;;
open Tcenv;;
open Config;;

(* Some global variables for interpreter state *)
let version = "0.3";;
let interactive = ref false;;
let verbose = ref false;;
let errors = ref 0;;
let parse_libs_var s = split (fun c -> c = ':') s;;
lib_dirs := 
   try (parse_libs_var (Sys.getenv "APROLOG_LIBS"))@ !lib_dirs
   with _ -> !lib_dirs

let add_lib_dir s = 
  lib_dirs := s::!lib_dirs
;;

let set_depth_first_bound i = 
  Flags.depth_first_bound := Some i;;

let timer = ref 0.0;;
let start_timer () = 
  timer := Sys.time();;
let time_msg msg = 
  let diff = Sys.time() -. !timer in
  print_string (msg^" took "^(string_of_float diff)^" s\n");
  flush stdout
;;


let find_local file = 
  if Sys.file_exists file then Some file else None
;;


let find_lib file = 
  let rec find l = 
    match l with 
      [] -> None
    | d::ds -> 
	if Sys.file_exists (d^"/"^file) 
	then Some (d^"/"^file) 
	else find ds
  in find !lib_dirs
;;



let quit() = 
  print_string "\nGoodbye.\n";
  exit(0)
;;


exception Success;;


let succeed() = raise Success;;


let lexbuf = Lexing.from_channel stdin;;


let get_lineno () = 
  Lineno.lineno()
;;

let parse file = 
  let ic = open_in file in
  Lineno.reset();
  Lineno.enter_file file;
  Parser.parse Lexer.lex (Lexing.from_channel ic)
;;


let parse_input_line () = 
  Lineno.reset();
  Parser.parse_input_line Lexer.lex lexbuf
;;


let get_decls find_file file = 
  match find_file file with 
    Some file' -> 
      let decls = 
	try parse file'
	with Parsing.Parse_error -> 
	  Util.error("Syntax error near ["^Lineno.filename()^":"
		       ^(string_of_int (get_lineno()))^"]\n")
	| Lexer.LexFail msg -> 
	    Util.error("Error near line ["^Lineno.filename()^":"
			 ^(string_of_int (get_lineno()))^"]: "^msg^"\n") 
      in
      decls
  |  None -> print_string (file^": File not found\n"); []
;;

(* TODO: Make the constrs data structure be like this intermediate 
   data structure *)
let simplify fvs (subst,constrs,evconstrs) = (subst,constrs,evconstrs)
(*  let rec mk_map constrs = 
    match constrs with
      [] -> (Varmap.empty)
    | (t,v)::constrs -> 
      let (vmap) = mk_map constrs in 
      let set = try Varmap.find v vmap with Not_found -> Varset.empty in
      (Varmap.add v (Varset.add t set) vmap)
    |  _ -> Util.impos "unexpected case in simplify"
  in let mk_constrs (vmap) = 
    let vss = Varmap.to_list vmap in
    let vls = List.map (fun (v,tset) -> (v,Varset.elements tset)) vss in
    let constrs = List.concat (List.map (fun (v,tset) -> List.map (fun t -> (t,v)) tset) vls) in
    constrs
  in (subst,mk_constrs(mk_map constrs))
;;
*)

let handleTcFail pos msg = 	
  (match pos with 
    None -> print_string ("***TYPE ERROR***\n"^msg^"\n")
  | Some p -> print_string ("***TYPE ERROR*** ["
			    ^Pos.pos2s p
			    ^"]: \n"^msg^"\n")
  );
  incr errors
;;


let help() = 
  let doc = text "Interpreter directives:" <:> newline
      <:> text "#quit.\t\t\tQuit AlphaProlog" <:> newline
      <:> text "#help.\t\t\tPrint this message" <:> newline
      <:> text "#use \"<filename>\".\tUse (read & execute) an external file" <:> newline
      <:> text "#open <id>.\t\tOpen a namespace (binding all its names locally)" <:> newline
      <:> text "#type <exp>.\t\tTypecheck an expression and display principal typing" <:> newline
      <:> text "#trace <num>.\t\tSet trace level to num = 0 (silent) to 3 (verbose)"<:> newline
      <:> text "#infix<dir> <id> <num>.\tMake identifier infix with associativity <dir> (l,r,n)"<:> newline
      <:> text "\t\t\tand precedence <num> (1..9)"<:>newline <:> newline
      <:> text "Keystroke commands:" <:> newline
      <:> text "<Control-C>\t\tInterrupt execution" <:> newline 
      <:> text "<Control-D>\t\tQuit AlphaProlog from prompt" <:> newline <:> newline
      <:> text "Queries:"<:> newline
      <:> text "<query>.\t\tPose a query." <:> newline<:> newline
(*      <:> text "Kinds of queries"<:> newline
      <:> text "<query>,<query> \tConjunction (\"and\") of queries" <:> newline
      <:> text "<query>;<query> \tDisjunction (\"or\") of queries" <:> newline
      <:> text "<exp> = <exp>\t\tUnification query" <:> newline
      <:> text "<exp> # <exp>\t\tFreshness query" <:> newline
      <:> text "<exp>\t\t\tAtomic query (e.g. \"List.append [1] [2] X\")"<:> newline <:> newline
*)
  in
  Printer.doc_to_channel stdout doc
  

let check_disjoint_vars vs = 
  let rec chk s vs = 
    match vs with 
      [] -> ()
    | v::vs -> 
	if Varset.mem v s 
	then raise (Tc.TcFail ("Repeated variable "^Var.to_string' v^" in type abbreviation"))
	else chk (Varset.add v s) vs
  in chk Varset.empty vs
;;


let rec run1 pos decl sg idx = 
  Var.reset_var();
  match decl with
    Query (t) as decl -> 
      let (tcenv,g) = Monad.run sg empty_env (Tc.check_goal t) in
      if not(!tc_only) && (!errors = 0 || !interactive)
      then (
	if !debug || not(!interactive)
	then (print_string "--------\nQuery: ";
	      Printer.print_to_channel Absyn.pp_term g stdout;
	      print_string "\n";
	      flush stdout
	      );
	let rec init_sc fvs ans = 
	  print_string "Yes.\n"; 
	  let ans = simplify fvs ans in
	  Printer.print_to_channel (Unify.pp_answer sg tcenv fvs) ans stdout;
	  flush stdout;
	  if !interactive 
	  then (
	    let s = input_line stdin in
	    if s = ";" then () else succeed())
	  else succeed()
	in
	(try 
	  let g' = Translate.translate_goal sg tcenv g in
	  if !debug 
	  then (
	    Printer.print_to_channel (Internal.pp_goal Isym.pp_term_sym) 
	      g' stdout;
	    print_string "\n"
	   );
	  let fvs = Varset.elements (Internal.fvs_g g') in
          Solve.solve Isym.pp_term_sym idx g' (init_sc fvs);
	  (* if we reach here, then no solution ws found *)
	  print_string "No.\n"
	with Success -> ()
	|  Sys.Break -> print_string "\nInterrupted.\n"));
      sg,idx
  | ClauseDecl(c) as decl -> 
      let (tcenv,p) = Monad.run sg empty_env (Tc.check_prog c) in
      if not(!tc_only) && (!errors = 0 || !interactive)
      then 
	(let p' = Translate.translate_prog sg tcenv p in
	let fas = Varset.elements(Internal.fas_p p') in
	let fvs = Varset.elements(Internal.fvs_p p') in
	let p' = List.fold_right (fun v p -> Internal.Dforall (v,p)) fvs p' in
	let p' = List.fold_right (fun a p -> Internal.Dnew(a,p)) fas p' in
	if !debug 
	then (
	  Printer.print_to_channel (Internal.pp_prog Isym.pp_term_sym) 
	    p' stdout;
	  print_string "\n"
	 );
	Index.add idx p');
      sg,idx
  | NamespaceDecl(sym,decls) -> 
      if !interactive || !verbose
      then (print_string ("namespace "^sym^" (\n"));
      Nstbl.enter_ns sg.tsg sym;
      Nstbl.enter_ns sg.ksg sym;
      let sg' = run sg idx decls in
      Nstbl.exit_ns sg.tsg;
      Nstbl.exit_ns sg.ksg;
      if !interactive || !verbose 
      then print_string ").\n";
      sg',idx
  | UseDirective(file) -> 
      let old_interactive = !interactive in
      interactive := false;
      let decls = get_decls find_lib file in
      let sg' = run sg idx decls in
      interactive := old_interactive;
      sg',idx
  | TraceDirective(i) -> 
      trace := i;
      sg,idx
  | KindDecl(s,k) -> 
      if !interactive || !verbose
      then (Printer.print_to_channel pp_decl {pos=pos;rdecl=decl} stdout;
	    print_string "\n");
      let u = Unique.get() in
      Nstbl.add sg.ksg (Nstbl.Rel(Nstbl.Base(s))) (k,u,None);
      let s' = Nstbl.resolve sg.ksg (Nstbl.Rel (Nstbl.Base s)) in
      Hashtbl.add sg.stbl u s';
      sg,idx
  | SymDecl (f,ty) -> 
      if !interactive || !verbose
      then (Printer.print_to_channel pp_decl {pos=pos;rdecl=decl} stdout;
	    print_string "\n");
      let (tcenv,ty) = Monad.run sg empty_env (Tc.check_ty ty false DataKd) in
      let u = Unique.get() in
      let entry = {is_def=false;sym_id = u;sym_ty = ty} in
      Nstbl.add sg.tsg (Nstbl.Rel(Nstbl.Base(f))) entry;
      let f' = Nstbl.resolve sg.tsg (Nstbl.Rel (Nstbl.Base f)) in
      Hashtbl.add sg.stbl u f';
      sg, idx
  | PredDecl(f,tys) -> 
      if !interactive || !verbose
      then (Printer.print_to_channel pp_decl {pos=pos;rdecl=decl} stdout;
	    print_string "\n");
      let ty = List.fold_right (fun t u -> ArrowTy (t,u)) tys PropTy in
      let (tcenv,ty) = Monad.run sg empty_env (Tc.check_ty ty false TypeKd) in
      let u = Unique.get() in
      let entry = {is_def=false;sym_id = u;sym_ty = ty} in
      Nstbl.add sg.tsg (Nstbl.Rel(Nstbl.Base(f))) entry;
      let f' = Nstbl.resolve sg.tsg (Nstbl.Rel (Nstbl.Base f)) in
      Hashtbl.add sg.stbl u f';
      sg, idx
      
  | TypeDefn (s,vs,ty) as decl -> 
      if !interactive || !verbose
      then (Printer.print_to_channel pp_decl {pos=pos;rdecl=decl} stdout;
	    print_string "\n");
      check_disjoint_vars vs;
      let ((_,tctx,_,_),ty) = Monad.run sg empty_env 
	  (Tc.check_ty ty true TypeKd) in
      let ks = 
	try List.map (fun v -> Varmap.find v tctx) vs 
	with Not_found -> 
	  raise (Tc.TcFail "Abbreviation has unconstrained type variables") in
      let ks = List.map (function NameKd -> NameK 
	                 |TypeKd|DataKd -> TypeK) ks in
      let k = List.fold_right (fun k k' -> ArrowK(k,k')) ks TypeK in
      let u = Unique.get() in
      Nstbl.add sg.ksg (Nstbl.Rel(Nstbl.Base(s))) 
	(k,u,Some (Tc.reflect_ty vs ty));
      let s' = Nstbl.resolve sg.ksg (Nstbl.Rel (Nstbl.Base s)) in
      Hashtbl.add sg.stbl u s';
      sg,idx
  | FuncDecl(f,tys,ty) as decl -> 
      if !interactive || !verbose
      then (Printer.print_to_channel pp_decl {pos=pos;rdecl=decl} stdout;
	    print_string "\n");
      let ty = List.fold_right (fun t u -> ArrowTy (t,u)) tys ty in
      let (_,ty) = Monad.run sg empty_env (Tc.check_ty ty false TypeKd) in
      let u = Unique.get() in
      let entry = {is_def=true;sym_id = u;sym_ty = ty} in
      Nstbl.add sg.tsg (Nstbl.Rel(Nstbl.Base(f))) entry;
      let f' = Nstbl.resolve sg.tsg (Nstbl.Rel (Nstbl.Base f)) in
      Hashtbl.add sg.stbl u f';
      sg,idx
  | BuiltinFuncDecl (sym,tys,ty,f) as decl -> 
      if !interactive || !verbose
      then (Printer.print_to_channel pp_decl {pos=pos;rdecl=decl} stdout;
	    print_string "\n");
      let ty = List.fold_right (fun t u -> ArrowTy (t,u)) tys ty in
      let (_,ty) = Monad.run sg empty_env (Tc.check_ty ty false TypeKd) in
      let u = Unique.get() in
      let entry = {is_def=false;sym_id = u;sym_ty = ty} in
      Nstbl.add sg.tsg (Nstbl.Rel(Nstbl.Base(sym))) entry;
      let sym' = Nstbl.resolve sg.tsg (Nstbl.Rel (Nstbl.Base sym)) in
      Hashtbl.add sg.stbl u sym';
      Runtime.bind u f;
      sg,idx
  | InfixDecl (_,_,_) -> sg,idx
  | TypeQuery (t) -> 
      let ((_,tctx,ctx,_),(t',ty)) = 
	Monad.run sg empty_env (Tc.infer_term t) in
      Printer.print_to_channel (Varmap.pp_map Tcenv.pp_symk) ctx stdout;
      print_string " |- ";
      Printer.print_to_channel pp_term t' stdout;
      print_string " : ";
      Printer.print_to_channel pp_ty ty stdout;
      nl();
      sg, idx
  | QuitDirective -> quit()
  | OpenDirective s -> 
      (try 
	Nstbl.open_ns sg.ksg s;
	Nstbl.open_ns sg.tsg s;
      with Not_found -> print_string ("Unknown namespace "^Nstbl.p2s s)
      );
      sg,idx
  | HelpDirective(_) -> (* ignore argument for now *)
      help();
      sg,idx


and run sg idx rest = 
  match rest with 
    [] -> sg
  | decl::decls -> 
      let sg,idx = 
	try run1 decl.pos decl.rdecl sg idx 
	with Tc.TcFail msg -> (
	  handleTcFail decl.pos msg;
	  sg,idx
	 ) 
      in
      run sg idx decls
;;


let rec toploop sg idx = 
  print_string "?- ";
  flush stdout;
  try 
    start_timer();
    let decls = parse_input_line() in
    if !verbose then time_msg("Parsing");
    start_timer();
    let sg = run sg idx decls in
    if !verbose then time_msg("Running");
    toploop sg idx
  with 
    Parsing.Parse_error -> 
      print_string ("I don't understand.  (Try \"#help.\")\n");
      toploop sg idx
  | Lexer.LexFail msg -> 
      print_string ("I don't understand.  (Try \"#help.\")\n");
      toploop sg idx 
  | End_of_file -> 
      quit()
  | Util.Impos msg -> 
      print_string ("Internal error: "^msg^".\n"^
		    "(This is my fault, not yours.  "^
		    "Please submit a bug report so I can fix the problem).\n");
      toploop sg idx
  | Error msg -> 
      print_string ("I don't understand.\n"^msg^".\n");
      toploop sg idx
  | RuntimeError msg -> 
      print_string ("A runtime error occurred:\n"^msg^".\n");
      toploop sg idx
  | Sys.Break -> (* No effort at atomicity here. *)
      print_string "\nInterrupted.\n";
      toploop sg idx
;;


let main args = 
  (* Banner hard to read because of escapes ... *)
  (* Looks like this
          ______                   
          ||   \\    /|           
          ||    )    ||           
    __  _ ||___/_ __ ||  __   __  _
   / \\/  || \\/ / \\|| / \\ / \\/
  (   )\  ||  ||(   )||(   )(   ) 
  \\_/ \\_||  ||\\_/ ||\\_/ \\_/  
          /   /      /      ((__ 
                            \\__))  
     *)
let banner = 
    "          ______                  \n"^
    "          ||   \\\\    /|          \n"^
    "          ||    )    ||           \n"^
    "    __  _ ||___/_ __ ||  __   __  _\n"^
    "   / \\\\/  || \\\\/ / \\\\|| / \\\\ / \\\\/\n"^
    "  (   )\\  ||  ||(   )||(   )(   ) \n"^
    "  \\\\_/ \\\\_||  ||\\\\_/ ||\\\\_/ \\\\_/  \n"^
    "          /   /      /      ((__ \n"^
    "                            \\\\__))  \n"
  in print_string banner;
  
  Sys.catch_break true; (* ctrl-C will raise Break *)
  let rsg = ref (Tcenv.init_sg()) in
  let idx = Index.create 100 in
  (* Initialize by interpreting runtime library decls *)
  rsg := run !rsg idx Runtime.decls;
  let do_file file = 
    print_string ("Reading file "^file^"...\n");
    start_timer();
    let decls = 
      try 
	get_decls find_local file 
      with Error msg -> print_string msg; exit(1)
    in
    if !verbose then time_msg "Parsing";
    start_timer();
    let sg = run !rsg idx decls in
    rsg := sg;
    if !verbose then time_msg "Running"
  in 
  let spec = [ ("-b", Arg.Int(set_depth_first_bound), "Depth-first search bound");
	       ("-d", Arg.Set(debug), "Enable debugging messages");
	       ("-hh", Arg.Clear(horn_clauses_only), 
		"Permit hereditary Harrop programs (experimental)");
	       ("-q", Arg.Set(quit_early), 
		"Quit interpreting after reading all input");
	       ("-L", Arg.String(add_lib_dir), "Add directory to library search path");
	       ("-sf", Arg.Set(strict_new_freshness),
		"Strict name freshness checks (experimental)");
	       ("-so", Arg.Set(strict_new_occurs),
		"Strict name occurs checks (experimental)");
	       ("-ss", Arg.Clear(fast_search),
		"Slow search (might get more answers)");
	       ("-susp", Arg.Set(Flags.suspensions),
	        "Apply permutations as suspensions (experimental)");
	       ("-new", Arg.Set(Flags.new_solve), "Use new solver");
	       ("-t", Arg.Int(fun i -> trace := i), "Set tracelevel (0,1,2)");
	       ("-to", Arg.Set(tc_only), "Typecheck only");
	       ("-v", Arg.Set(verbose), "Verbose mode");
	     ]
  in
  print_string ("AlphaProlog "^version^"\n");
  print_string ("For help, type \"#help.\"\n");
  flush stdout;
  Arg.parse spec do_file "usage: aprolog [options] files";
  if !errors = 0 && not (!quit_early)
  then (interactive := true;
	toploop !rsg idx)
  else (if !errors != 0 
        then  (print_string (string_of_int !errors ^ " errors\n");
	       exit(1)))
;;


main Sys.argv;;
    
