(* Alpha Prolog *)
open Types;;
open Internal;;
module P = Perm;;
module S = Subst;;
open Printer;;
    
(* TODO: Use a faster version of unification such as using reference cells *)
    
(* TODO: Implement "complete" unification for arbitrary terms. *)
    
(* TODO: Interleave freshness constraint solving with unification whenever a 
    substitution is performed (solve constraints "eagerly" *)
    
type 'a constrs = ('a term * var) list;;


type 'a evconstrs = (('a term * 'a term) list) list;;


type 'a fprobs = ('a term * 'a term) list;;


type 'a answer = ('a S.subst * 'a constrs * 'a evconstrs);;


(* return value indicates whether failure was due to mismatched atoms or
   some other reason: this is used to determine whether to try a
   different atom instantiation *)
type 'a sc = 'a -> unit;;
type 'a sc_transform = 'a sc -> 'a sc;;


let new_var(x) = Susp(Perm.id,Exist Toplevel,Var.rename x);;


let rec compress t sub = 
  match t with
    Susp(p,Exist b,x) -> 
      (match S.lookup sub x with 
	  None -> t
	| Some t' -> compress (apply_p p t') sub)
  | t -> t
;;


let rec unify_fresh1 t1 t2 sub sc constrs = 
  match compress t1 sub, compress t2 sub with
  | Name a,Name b -> 
      if Var.eq a b 
      then ()
      else sc constrs
  | Name a,Abs(b,e) -> 
      if Var.eq a b 
      then sc constrs
      else 
	unify_fresh1 t1 e sub sc constrs
  | Susp(p,b,v), Abs (a,t) -> 
      let c = Var.rename a in
      unify_fresh1 t1 (apply_p (P.trans a c) t) sub 
        (fun constrs -> 
          sc (List.filter (fun (t',_) -> not(t' = Name c)) constrs))
        constrs
  | _,App(c,es) -> 
      unifys_fresh (List.map (fun t2 -> (t1,t2)) es) sub sc constrs
  | Susp(p,b,x), Name a 
  | Name a, Susp(p,b,x) ->  
      let a' = (P.lookup (P.inv p) a) in
      sc ((Name a', x)::constrs)
  | Susp(p,b,x), Susp(q,c,y) -> 
    let pi' = Perm.comp (Perm.inv q) p in
    if P.eq p q && Var.eq x y 
    then ()
    else 
      (sc ((Susp(pi', b,x), y)::constrs))
  | _,_ -> Util.impos "illegal case in freshness unification"

and unifys_fresh fs sub sc constrs = 
  match fs with
    [] -> sc constrs 
  | (a,e)::fs -> unify_fresh1 a e sub (unifys_fresh fs sub sc) constrs
;;


let do_susp p1 p2 x sc (sub,fs,es) = 
  let ds = P.disagreement_set p1 p2 in
  let new_fs = List.map (fun a -> (Name a,x)) ds in
  sc (sub, new_fs@fs,es)
;;


let bconstr_leq c1 c2 = 
  match c1,c2 with
    _,Toplevel -> true
  | Constrained l, Constrained l' ->
   List.length l <= List.length l'
  | _,_ -> false
;;


let bconstr_satisfies x c = 
  match c with
    Toplevel -> true
  | Constrained l -> List.exists (Var.eq x) l
;;


(* Extended occurs check: check that x does not occur in t, that
   all Uvars in t are permitted for x, and constrain all Evars in 
   t to be as constrained as x is.  *)
let rec occurs_check x c t sc subst = 
  match compress t subst with
    Name a -> 
      if !Flags.strict_new_occurs
      then 
        if bconstr_satisfies a c
        then sc (t,subst)
        else ()
      else sc (t,subst)
  | Abs(a,e) -> 
      occurs_check x c e (fun (t',subst) -> 
      sc (Abs(a,t'), subst)) subst
  | App(f,es) -> 
      occurs_checks x c es (fun (ts,subst) -> 
      sc (App(f,es),subst)) subst
  | Susp(p,Univ,y) -> 
      if bconstr_satisfies y c
      then sc (t,subst)
      else ()
  | Susp(p,Exist c',y) -> 
      if Var.eq x y 
      then ()
      else 
        if bconstr_leq c' c (* If lhs more constrained than rhs *)
	then sc (t,subst) (* continue *)
	else (sc (Susp(p,Exist c, y),subst)) 
and occurs_checks x c ts sc subst = 
  match ts with
    [] -> sc ([],subst)
  | t::ts -> 
      occurs_check x c t (fun (t',subst) -> 
      occurs_checks x c ts (fun (ts',subst) -> 
      sc (t'::ts',subst)) subst) subst
;;

let is_relevant_ev x (ev:('a term * 'a term) list) = 
  let is_relevant e = 
  match e with 
    (Susp(_,_,y),Susp(_,_,y')) -> Var.eq x y || Var.eq x y'
  | (Susp(_,_,y),u) -> Var.eq x y
  | (t,Susp(_,_,y)) -> Var.eq x y
  | _ -> false
  in List.exists is_relevant ev
;;


let rec is_relevant_f x f = 
  match f with 
    (Susp(_,_,y),y') as f -> Var.eq x y || Var.eq x y' 
  | (t,y') -> Var.eq x y' 
;;
  

	
let rec unify_wrap eq t1 t2 sc ans = 
  let rec do_subst t p b x sc (subst,constrs,evconstrs) = 
    match S.lookup subst x with
      None -> 
	occurs_check x b t (fun (t',subst) -> 
	let subst' = S.join x (apply_p (P.inv p) t') subst in
        let revs, ievs = List.partition (is_relevant_ev x) evconstrs in 
        let relevant_evs = List.map Util.unzip revs in
        eunifys'' eq relevant_evs (fun (subst'', constrs, evconstrs') -> 
	let rfs,ifs = List.partition (is_relevant_f x) constrs in 
        let relevant_fs = List.map (fun (t,x) -> (t,Susp(P.id, Exist Toplevel, x))) rfs in
	unifys_fresh relevant_fs subst'' (fun constrs' -> 
	  sc (subst'',constrs',evconstrs')) ifs) (subst',constrs,ievs)) subst
    | Some t' -> 
	unify1 t (apply_p p t') sc (subst, constrs,evconstrs) 
	  
  and do_EE_subst (p1,c1,x) (p2,c2,y) sc ans = 
    do_subst (Susp(p2,Exist c2,y)) p1 c1 x sc ans 
      
  and unify1 t1 t2 sc ans = 
   let (subst,_,_) = ans in 
    match compress t1 subst, compress t2 subst with
      (Name a, Name b) -> 
	if Var.eq a b 
	then sc ans
	else ()
    | (Abs(a,e), Abs(b,e')) -> 
	if Var.eq a b 
	then unify1 e e' sc ans
	else 
	  unify1 e (apply_p (P.trans a b) e') 
	    (fun (s,c,e) -> unify_fresh1 (Name a) e' s (fun c' -> sc (s,c',e)) c) ans
    | (App(c,es),App(d,es')) -> 
	if eq c d 
	then unifys eq es es' sc ans
	else ()
    | (Susp(p1,Univ,x),Susp(p2,Univ,y)) -> 
	if Var.eq x y && P.disagreement_set p1 p2 = []
	then sc ans 
	else ()
    | (Susp(p1,Exist b1,x),Susp(p2,Exist b2,y)) -> 
	if Var.eq x y 
	then do_susp p1 p2 x sc ans
	else do_EE_subst (p1,b1,x) (p2,b2,y) sc ans
    | (Susp(p,Exist b,x), t) -> 
	do_subst t p b x sc ans
    | (t,Susp(p,Exist b,x)) -> 
	do_subst t p b x sc ans
    | _,_ -> ()
  in 
  unify1 t1 t2 sc ans
and unifys aeq es es' sc ans = 
    match es, es' with
      [],[] -> sc ans
    | e::es,e'::es' -> 
	unify_wrap aeq e e' (unifys aeq es es' sc) ans
    | _,_ -> ()

(* TODO: Keep iterating until no new substitutions occur. !! *)
and classify aeq t u sc classes ans = 
   let (subst,_,_) = ans in 
     match classes with 
       [] -> sc [(t,u)] ans
     | (t',u')::classes -> 
       if (Internal.eq aeq (compress t subst) (compress t' subst))
          || (Internal.eq aeq (compress u subst) (compress u' subst))
       then unifys aeq [t;u] [t';u'] (classify aeq t u sc classes) ans
       else classify aeq t u (fun classes -> 
              sc ((t',u')::classes)) classes ans 
(* TODO: Rewrite this to chomp off a maximal chunk of term and do 
   occurs check at the same time *)
and eunify' aeq  t u sc classes ans = 
   let (subst,fs,es) = ans in 
    match compress t subst, compress u subst with
    | App(f,ts'), App(f',us') -> 
        if aeq f f' 
        then eunifys' aeq ts' us' sc classes ans
    | Susp(_,Exist c,x), App(f,us') -> 
        let vs = List.map (fun _ -> new_var(x)) us' in 
	occurs_check x c u (fun (_,subst) -> unify_wrap aeq (App(f,vs)) t (eunifys' aeq vs us' sc classes) (subst,fs,es)) subst
    | App(f,ts'),Susp(_,Exist(c),y) -> 
	let vs = List.map (fun _ -> new_var(y)) ts' in 
	occurs_check y c t (fun (_,subst) -> unify_wrap aeq (App(f,vs)) u (eunifys' aeq ts' vs sc classes) (subst,fs,es)) subst
    | Abs(a,t'),Abs(b,u') -> 
        let c = Var.rename a in
        let t'' = apply_p (Perm.trans a c) t' in
	let u'' = apply_p (Perm.trans b c) u' in
	eunifys' aeq [Name(c);t''] [Name(c);u''] sc classes ans
    | Abs(a,t),Susp(_,Exist(con),y) -> 
	let c = Var.rename a in
        let t' = apply_p (Perm.trans a c) t in
	let z = new_var(y) in 
	occurs_check y con t (fun (_,subst) -> 
          unify_wrap aeq (Abs(c,z)) u (
	  eunifys' aeq [Name(c);t'] [Name(c);z] sc classes) (subst,fs,es)) subst    | Susp(_,Exist(con),x),Abs(b,u) -> 
	let c = Var.rename b in
	let z = new_var(x) in 
        let u' = apply_p (Perm.trans b c) u in
        occurs_check x con u (fun (_,subst) -> 
          unify_wrap aeq (Abs(c,z)) t (
	  eunifys' aeq [Name(c);z] [Name(c);u'] sc classes) (subst,fs,es)) subst
    | _,_ -> classify aeq t u sc classes ans
and eunifys' aeq ts us sc classes ans = 
    match ts,us with
      [],[] -> sc classes ans
    | t::ts',u::us' -> eunify' aeq t u (eunifys' aeq ts' us' sc) classes ans
    | _,_ -> ()
and eunifys'' aeq tus sc ans = 
   match tus with 
     (ts,us)::tus -> eunifys' aeq ts us (fun classes (subst,fs,es) -> 
        let es = if List.length classes <= 1 then es else classes::es
   in 
    eunifys'' aeq tus sc (subst,fs,es)) [] ans
   | [] -> sc ans
;;

let eunify aeq t u sc ans =   
  eunify' aeq t u (fun classes (subst,fconstrs,evconstrs) -> 
   let es = if List.length classes <= 1 then evconstrs else classes::evconstrs
   in 
    sc (subst,fconstrs,es)) [] ans
;;


let unify_fresh a t sc (sub,constrs,evconstrs) =
   unify_fresh1 a t sub (fun constrs' -> sc (sub,constrs',evconstrs)) constrs
;;


let unify eq e1 e2 sc (sub,constrs,evconstrs) = 
  unify_wrap eq e1 e2 sc (sub,constrs,evconstrs)
;;


let rec unifys eq es sc = 
  match es with
    [] -> sc
  | (t1,t2)::es -> unify eq t1 t2 (unifys eq es sc)
;;


let rec pp_constrs pp constrs = 
  match constrs with
    [] -> emp
  | (t,x)::constrs -> 
      Internal.pp_term pp t <+>
      text "#" <+>
      text (Var.to_string' x) <:>
      newline <:>
      pp_constrs pp constrs
;;

	
let pp_evconstrs pp evcs = 
  let pp_evconstr evc =
    let lc,rc = Util.unzip evc in 
    paren (sep comma (List.map pp lc)) 
      <+> tilde <:> eq <+> 
    paren (sep comma (List.map pp rc))
  in 
  vcat (List.map pp_evconstr evcs)


let rec pp_answer sg env fvs (subst,constrs,evconstrs) = 
  match fvs with
   [] -> 
     pp_constrs Isym.pp_term_sym constrs <:> newline <:> 
     pp_evconstrs (S.pp_term_subst Isym.pp_term_sym subst) evconstrs
   | x::xs -> 
     match S.finish subst x with
       None -> pp_answer sg env xs (subst,constrs,evconstrs) 
     | Some t -> 
         let t' = Translate.untranslate_tm sg env t in 
         text (Var.to_string' x) <+> 
         text "=" <+>
         Absyn.pp_term t' <+>
         newline <:>
         pp_answer sg env xs (subst, constrs,evconstrs)
;;
